
namespace Softshell
{
  __constant__ OctreeEntry* octree;
  __constant__ unsigned int *vertexIdOut, *triOutput;
  __constant__ const unsigned int *vertexCoherence, *triangleCoherence;
  __constant__ unsigned int vertices, maxdepth;
  __constant__ const uint3* voxelIds;

  __device__ uint triangleOutCount;


  typedef float Params;
  typedef uint2 NodeWorkItem;
  typedef CombWorkpackage<NodeWorkItem, Params> NodeWp;


  class GenVerticesProcedure: public Procedure
  {
  public:
    __device__ static void execute(volatile Workpackage* workpackage)
    {
      volatile NodeWp* mywp = static_cast<volatile NodeWp*>(workpackage);

      uint mynodeid = mywp->getMyWorkItem().x;
      uint mynodedepth = mywp->getMyWorkItem().y;

      const OctreeEntry* node = octree + mynodeid;

      uint vertexTargetStart = node->vertexOffset;
      uint vertexTargetEnd = vertexTargetStart + node->numVertices;
      bool leafnode = mynodedepth >= maxdepth;
      for(uint targetVertex = vertexTargetStart; targetVertex < vertexTargetEnd; ++targetVertex)
      {
        uint coherenceId = vertexCoherence[targetVertex];
        vertexIdOut[coherenceId] = leafnode?coherenceId:(vertices + mynodeid);
      }

      if(__threadId() == 0)
        delete mywp;
    }
  };

  template<uint DepthError>
  class TraverseProcedure : public Procedure
  {
    __device__ static void  writeTriangleOutput(const OctreeEntry* node )
    {
      __shared__ uint myshared[BlockSize+1];
      uint triOffset = 0, numTris = 0;
      if(node != 0)
      {
        triOffset = node->triangleOffset;
        numTris = node->numTriangles;
      }
      myshared[__threadId()] = numTris;
      //run prefix sum to get local offset
      scan_workefficient(myshared, BlockSize);
      if(__threadId() == BlockSize-1)
        myshared[BlockSize] = atomicAdd(&triangleOutCount, myshared[__threadId()]+numTris);
      __sync();

      if(numTris == 0)
        return;

      uint startWrite = myshared[BlockSize] + myshared[__threadId()];

      numTris += triOffset;
      uint * indexoutput = triOutput + 3*startWrite;


      //run through all triangles
      for(uint i = triOffset; i < numTris; ++i)
      {
        //all 3 vertices
        for(unsigned int v = 0; v < 3; ++v)
          *indexoutput++ = vertexIdOut[triangleCoherence[3*i + v]];
      }
    }

  public:

    __device__ static void execute(volatile Workpackage* workpackage)
    {
      volatile NodeWp* mywp = static_cast<volatile NodeWp*>(workpackage);

      const float threshold = *mywp->params;
      const OctreeEntry* node = 0;

      if(mywp->hasWork())
      {
        uint mynodeid = mywp->getMyWorkItem().x;
        uint mynodedepth = mywp->getMyWorkItem().y;
       
        uint childDepth = mynodedepth + 1;
        const OctreeEntry* node = octree + mynodeid;

        //check children to be expanded
        uint numChildren = OctreeEntry::hasChildren(node->children)?OctreeEntry::getNumChildren(node->children):0;
        uint firstChildId = OctreeEntry::getFirstChild(node->children);
        for(uint i = 0; i < numChildren; ++i)
        {
          uint thisChildId = firstChildId + i;
          const OctreeEntry *childEntry = octree + thisChildId;
          float3 vMin, vMax;
          getVoxelBounds(voxelIds[thisChildId], childDepth, vMin, vMax);
          //check if criterion is fullfilled
          bool expand = carryOnTraversal<DepthError>(childDepth, vMin, vMax, threshold);
          if(expand && childDepth < maxdepth)
            issueWorkItem<TraverseProcedure>(make_uint2(thisChildId, childDepth), mywp->params);
          else //leaf
            issueWorkItem<GenVerticesProcedure>(make_uint2(thisChildId, childDepth), mywp->params);
        }
        if(numChildren == 0) //add vertices for this node
          issueWorkItem<GenVerticesProcedure>(make_uint2(thisChildId, childDepth), mywp->params);
      }

      writeTriangleOutput(node);

      __sync();

      if(__threadId() == 0)
        delete mywp;
    }
    static const uint BlockSize = 256;
  };

  //Priority class telling the system to execute TraverseProcedures first
  class TraversePriority
  {
  public:
    static __device__ float eval(Workpackage* workpackage, ProcedureInfo* proc)
    {
      //note: the workpackage wp will be called for the procedue proc
      if(procEqual<TraverseProcedure>(*proc))
        return 1.0f;
      else
        return 0.0f;
    }
  };

  template<uint DepthError>
  class TraverseEvent : public EventCreator
  {
  public:
    __device__ static void create(Params* params, void* &storage)
    {
      NodeWorkItem root = make_uint2(0,0);
      emitWorkpackage<TraverseProcedure<DepthError> >(new NodeWp(&root, 1, params));
    }
  };

};