#ifndef COREFUNCTIONS_CUH
#define COREFUNCTIONS_CUH

#include "structures.h"

__device__ float device_random( uint& seed )
{
  const uint MULTIPLIER = 1664525;
  const uint OFFSET = 1013904223;
  const double MODULUS = 4294967295.0;
  const float MODULUS_INV = (float) (1.0 / MODULUS);

	seed = seed * MULTIPLIER + OFFSET;
	float res = seed * MODULUS_INV;
	return res;
}

__device__ bool intersectSphere(const object& sphere, float3 &d, float3 &n, float& mindist, const float3& orig, const float3& dir)
{
  d = sphere.pos - orig;
  float v = vec_dot( dir, d );

  if( v - sphere.rad > mindist )
    return false;

  float t = sphere.rad*sphere.rad + v*v - d.x*d.x - d.y*d.y - d.z*d.z;
  if( t < 0 )
    return false;

  t = v - sqrt( t );
  if( ( t > mindist ) || ( t < 0 ) )
    return false;

  n = orig + t*dir - sphere.pos;
  n = vec_normalize( n );

  mindist = t;
  return true;
}

__device__ bool intersectPlane(const object& plane, vector3 &d, vector3 &n, float& mindist, const vector3& orig, const vector3& dir)
{
  float v = vec_dot( plane.n, dir );

	if( v >= 0 )
		return false;

  d = plane.pos - orig;

	float t = vec_dot( plane.n, d ) / v;
	if( ( t > mindist ) || ( t < 0 ) )
		return false;

	n = plane.n;
	mindist = t;
  return true;
}

__device__ void hitObject(const object& hitObject, float t, const vector3 &n, vector3& orig, vector3& dir, vector3& contrib, rgbcolor& color, uint& seed)
{
  const float EPSILON = 0.001f;
  vector3 hit = orig + t*dir;

	t = 2.0f * vec_dot( -dir, n );
  dir = t*n+dir;
  dir = vec_normalize( dir );
				
  vector3 o;
	do
	{
    o = 2.0f*make_float3(device_random(seed),device_random(seed),device_random(seed))-1.0f;
	}
	while( ( vec_dot(o,o) > 1 ) || ( vec_dot( o, n ) <= 0 ) );

	float v = (hitObject.d*device_random(seed))*hitObject.g + hitObject.d*(1.0f-hitObject.g);
  dir = o*v + dir*(1.0f-v);
  dir = vec_normalize( dir );

  orig = hit + EPSILON*dir;

  contrib *= hitObject.c;

  color += hitObject.e * contrib;
}

__device__ const object* intersectObjects(const object* objects, uint numobjects, float3 &d, float3 &n, float& t, const float3& orig, const float3& dir)
{
  t = 1000000.0f;
  const object* res = 0;
  for(uint i = 0; i < numobjects; ++i)
  {
    if(objects[i].type == TYPE_SPHERE)
    {
      if(intersectSphere(objects[i], d, n, t, orig, dir))
        res = objects + i;
    }
    else if(objects[i].type == TYPE_PLANE)
    {
      if(intersectPlane(objects[i], d, n, t, orig, dir))
        res = objects + i;
    }
  }
  return res;
}

#endif
